@@ -1,3 +1,4 @@
+{-# LANGUAGE TypeFamilies #-}
 -- | Module implementing GIF decoding.
 module Codec.Picture.Gif ( -- * Reading
                            decodeGif
@@ -39,6 +40,8 @@ import Data.Binary.Get( Get
                       , getWord8
                       , getWord16le
                       , getByteString
+                      , bytesRead
+                      , skip
                       )
 
 import Data.Binary.Put( Put
@@ -186,19 +189,17 @@ data ImageDescriptor = ImageDescriptor
   , gDescLocalColorTableSize    :: !Word8
   }
 
-imageSeparator, extensionIntroducer, gifTrailer, applicationLabel :: Word8
+imageSeparator, extensionIntroducer, gifTrailer :: Word8
 imageSeparator      = 0x2C
 extensionIntroducer = 0x21
 gifTrailer          = 0x3B
+
+graphicControlLabel, commentLabel, plainTextLabel, applicationLabel :: Word8
+plainTextLabel = 0x01
+graphicControlLabel = 0xF9
+commentLabel = 0xFE
 applicationLabel    = 0xFF
 
-graphicControlLabel :: Word8
-graphicControlLabel = 0xF9
-
---commentLabel, graphicControlLabel, 
---
-{-commentLabel        = 0xFE-}
-{-plainTextLabel      = 0x01-}
 
 parseDataBlocks :: Get B.ByteString
 parseDataBlocks = B.concat <$> (getWord8 >>= aux)
@@ -214,9 +215,31 @@ putDataBlocks wholeString = putSlices wholeString >> putWord8 0
         putSlices str =
             putWord8 (fromIntegral $ B.length str) >> putByteString str
 
+data DisposalMethod
+    = DisposalAny
+    | DisposalDoNot
+    | DisposalRestoreBackground
+    | DisposalRestorePrevious
+    | DisposalUnknown Word8
+
+disposalMethodOfCode :: Word8 -> DisposalMethod
+disposalMethodOfCode v = case v of
+    0 -> DisposalAny
+    1 -> DisposalDoNot
+    2 -> DisposalRestoreBackground
+    3 -> DisposalRestorePrevious
+    n -> DisposalUnknown n
+
+codeOfDisposalMethod :: DisposalMethod -> Word8
+codeOfDisposalMethod v = case v of
+    DisposalAny -> 0
+    DisposalDoNot -> 1
+    DisposalRestoreBackground -> 2
+    DisposalRestorePrevious -> 3
+    DisposalUnknown n -> n
 
 data GraphicControlExtension = GraphicControlExtension
-    { gceDisposalMethod        :: !Word8 -- ^ Stored on 3 bits
+    { gceDisposalMethod        :: !DisposalMethod -- ^ Stored on 3 bits
     , gceUserInputFlag         :: !Bool
     , gceTransparentFlag       :: !Bool
     , gceDelay                 :: !Word16
@@ -228,8 +251,9 @@ instance Binary GraphicControlExtension where
         putWord8 extensionIntroducer
         putWord8 graphicControlLabel
         putWord8 0x4  -- size
-        let disposalField =
-                (gceDisposalMethod v .&. 0x7) `unsafeShiftL` 2
+        let disposalCode = codeOfDisposalMethod $ gceDisposalMethod v
+            disposalField =
+                (disposalCode .&. 0x7) `unsafeShiftL` 2
 
             userInputField
                 | gceUserInputFlag v = 0 `setBit` 1
@@ -257,7 +281,9 @@ instance Binary GraphicControlExtension where
         idx              <- getWord8
         _blockTerminator <- getWord8
         return GraphicControlExtension
-            { gceDisposalMethod        = (packedFields `unsafeShiftR` 2) .&. 0x07
+            { gceDisposalMethod        = 
+                disposalMethodOfCode $
+                    (packedFields `unsafeShiftR` 2) .&. 0x07
             , gceUserInputFlag         = packedFields `testBit` 1
             , gceTransparentFlag       = packedFields `testBit` 0
             , gceDelay                 = delay
@@ -296,19 +322,34 @@ instance Binary GifImage where
 data Block = BlockImage GifImage
            | BlockGraphicControl GraphicControlExtension
 
+skipSubDataBlocks :: Get ()
+skipSubDataBlocks = do
+  s <- fromIntegral <$> getWord8
+  if s == 0 then
+    return ()
+  else
+    skip s >> skipSubDataBlocks
+
 parseGifBlocks :: Get [Block]
 parseGifBlocks = getWord8 >>= blockParse
-  where blockParse v
-          | v == gifTrailer = pure []
-          | v == imageSeparator = (:) <$> (BlockImage <$> get) <*> parseGifBlocks
-          | v == extensionIntroducer = do
-                extensionCode <- getWord8
-                if extensionCode /= graphicControlLabel
-                   then parseDataBlocks >> parseGifBlocks
-                   else (:) <$> (BlockGraphicControl <$> get) <*> parseGifBlocks
+  where
+    blockParse v
+      | v == gifTrailer = pure []
+      | v == imageSeparator = (:) <$> (BlockImage <$> get) <*> parseGifBlocks
+      | v == extensionIntroducer = getWord8 >>= extensionParse
+
+    blockParse v = do
+      readPosition <- bytesRead
+      fail ("Unrecognized gif block " ++ show v ++ " @" ++ show readPosition)
+
+    extensionParse code
+     | code == graphicControlLabel =
+        (:) <$> (BlockGraphicControl <$> get) <*> parseGifBlocks
+     | code == commentLabel = skipSubDataBlocks >> parseGifBlocks
+     | code `elem` [plainTextLabel, applicationLabel] =
+        fromIntegral <$> getWord8 >>= skip >> skipSubDataBlocks >> parseGifBlocks
+     | otherwise = parseDataBlocks >> parseGifBlocks
 
-        blockParse v = do
-            fail ("Unrecognized gif block " ++ show v)
 
 instance Binary ImageDescriptor where
     put v = do
@@ -365,7 +406,8 @@ instance Binary ImageDescriptor where
 ----            Palette
 --------------------------------------------------
 getPalette :: Word8 -> Get Palette
-getPalette bitDepth = replicateM (size * 3) get >>= return . Image size 1 . V.fromList
+getPalette bitDepth = 
+    replicateM (size * 3) get >>= return . Image size 1 . V.fromList
   where size = 2 ^ (fromIntegral bitDepth :: Int)
 
 putPalette :: Int -> Palette -> Put
@@ -394,7 +436,13 @@ instance Binary GifHeader where
     get = do
         version    <- get
         screenDesc <- get
-        palette    <- getPalette $ colorTableSize screenDesc
+        
+        palette <- 
+          if hasGlobalMap screenDesc then
+            getPalette $ colorTableSize screenDesc
+          else
+            return $ greyPalette
+
         return GifHeader
             { gifVersion = version
             , gifScreenDescriptor = screenDesc
@@ -501,18 +549,23 @@ decodeAllGifImages GifFile { gifImages = [] } = []
 decodeAllGifImages GifFile { gifHeader = GifHeader { gifGlobalMap = palette
                                                    , gifScreenDescriptor = wholeDescriptor
                                                    }
-                           , gifImages = (_, firstImage) : rest } = map paletteApplyer $
- scanl generator (paletteOf palette firstImage, decodeImage firstImage) rest
-    where globalWidth = fromIntegral $ screenWidth wholeDescriptor
+                           , gifImages = (firstControl, firstImage) : rest } = map paletteApplyer $
+ scanl generator initState  rest
+    where initState = (paletteOf palette firstImage, firstControl, decodeImage firstImage)
+          globalWidth = fromIntegral $ screenWidth wholeDescriptor
           globalHeight = fromIntegral $ screenHeight wholeDescriptor
 
-          {-background = backgroundIndex wholeDescriptor-}
+          background = backgroundIndex wholeDescriptor
+          backgroundImage = generateImage (\_ _ -> background) globalWidth globalHeight
 
-          paletteApplyer (pal, img) = substituteColors pal img
+          paletteApplyer (pal, _, img) = substituteColors pal img
 
-          generator (_, img1) (controlExt, img2@(GifImage { imgDescriptor = descriptor })) =
-                        (paletteOf palette img2, generateImage pixeler globalWidth globalHeight)
-               where localWidth = fromIntegral $ gDescImageWidth descriptor
+          generator (_, prevControl, img1)
+                    (controlExt, img2@(GifImage { imgDescriptor = descriptor })) =
+                        (thisPalette, controlExt, thisImage)
+               where thisPalette = paletteOf palette img2
+                     thisImage = generateImage pixeler globalWidth globalHeight
+                     localWidth = fromIntegral $ gDescImageWidth descriptor
                      localHeight = fromIntegral $ gDescImageHeight descriptor
 
                      left = fromIntegral $ gDescPixelsFromLeft descriptor
@@ -530,15 +583,24 @@ decodeAllGifImages GifFile { gifHeader = GifHeader { gifGlobalMap = palette
                             then fromIntegral $ gceTransparentColorIndex ext
                             else 300
 
+                     oldImage = case gceDisposalMethod <$> prevControl of
+                        Nothing -> img1
+                        Just DisposalAny -> img1
+                        Just DisposalDoNot -> img1
+                        Just DisposalRestoreBackground -> backgroundImage
+                        Just DisposalRestorePrevious -> img1
+                        Just (DisposalUnknown _) -> img1
+
                      pixeler x y
                         | isPixelInLocalImage x y && fromIntegral val /= transparent = val
                             where val = pixelAt decoded (x - left) (y - top)
-                     pixeler x y = pixelAt img1 x y
+                     pixeler x y = pixelAt oldImage x y
 
 decodeFirstGifImage :: GifFile -> Either String (Image PixelRGB8)
-decodeFirstGifImage
-        GifFile { gifHeader = GifHeader { gifGlobalMap = palette}
-                , gifImages = ((_, gif):_) } = Right . substituteColors palette $ decodeImage gif
+decodeFirstGifImage img@GifFile { gifImages = (firstImage:_) } =
+    case decodeAllGifImages img { gifImages = [firstImage] } of
+      [] -> Left "No image after decoding"
+      (i:_) -> Right i
 decodeFirstGifImage _ = Left "No image in gif file"
 
 -- | Transform a raw gif image to an image, witout
@@ -631,7 +693,7 @@ encodeGifImages looping imageList@((firstPalette, _,firstImage):_) = Right $ enc
 
     controlExtension 0 =  Nothing
     controlExtension delay = Just GraphicControlExtension
-        { gceDisposalMethod        = 0
+        { gceDisposalMethod        = DisposalAny
         , gceUserInputFlag         = False
         , gceTransparentFlag       = False
         , gceDelay                 = fromIntegral delay
