@@ -1,6 +1,14 @@
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE TypeFamilies #-}
 -- | Module dedicated of Radiance file decompression (.hdr or .pic) file.
 -- Radiance file format is used for High dynamic range imaging.
-module Codec.Picture.HDR( decodeHDR, encodeHDR, writeHDR ) where
+module Codec.Picture.HDR( decodeHDR
+                        , encodeHDR
+                        , encodeRawHDR
+                        , encodeRLENewStyleHDR
+                        , writeHDR
+                        , writeRLENewStyleHDR
+                        ) where
 
 import Data.Bits( Bits, (.&.), (.|.), unsafeShiftL, unsafeShiftR )
 import Data.Char( ord, chr, isDigit )
@@ -9,7 +17,6 @@ import Data.Monoid( (<>) )
 import Control.Applicative( pure, (<$>), (<*>) )
 import Control.Monad( when, foldM, foldM_, forM, forM_ )
 import Control.Monad.Trans.Class( lift )
-import Control.Monad.Trans.Error( ErrorT, throwError, runErrorT )
 import qualified Data.ByteString as B
 import qualified Data.ByteString.Lazy as L
 import qualified Data.ByteString.Char8 as BC
@@ -29,6 +36,21 @@ import Codec.Picture.InternalHelper
 import Codec.Picture.Types
 import Codec.Picture.VectorByteConversion
 
+#if MIN_VERSION_transformers(0, 4, 0)
+import Control.Monad.Trans.Except( ExceptT, throwE, runExceptT )
+#else
+-- Transfomers 0.3 compat
+import Control.Monad.Trans.Error( Error, ErrorT, throwError, runErrorT )
+
+type ExceptT = ErrorT
+
+throwE :: (Monad m, Error e) => e -> ErrorT e m a
+throwE = throwError
+
+runExceptT :: ErrorT e m a -> m (Either e a)
+runExceptT = runErrorT
+#endif
+
 {-# INLINE (.<<.) #-}
 (.<<.), (.>>.) :: (Bits a) => a -> Int -> a
 (.<<.) = unsafeShiftL
@@ -40,7 +62,7 @@ import Codec.Picture.VectorByteConversion
 (.<-.) = M.write 
          {-M.unsafeWrite-}
 
-type HDRReader s a = ErrorT String (ST s) a
+type HDRReader s a = ExceptT String (ST s) a
 
 data RGBE = RGBE !Word8 !Word8 !Word8 !Word8
 
@@ -144,9 +166,9 @@ decodeInfos = do
 --  * PixelRGBF
 --
 decodeHDR :: B.ByteString -> Either String DynamicImage
-decodeHDR str = runST $ runErrorT $ do
+decodeHDR str = runST $ runExceptT $ do
     case runGet decodeHeader $ L.fromChunks [str] of
-      Left err -> throwError err
+      Left err -> throwE err
       Right rez ->
           ImageRGBF <$> (decodeRadiancePicture rez >>= lift . unsafeFreezeImage)
 
@@ -214,7 +236,7 @@ newStyleRLE :: L.ByteString -> Int -> M.STVector s Word8
             -> HDRReader s Int
 newStyleRLE inputData initialIdx scanline = foldM inner initialIdx [0 .. 3]
   where dataAt idx
-            | fromIntegral idx >= maxInput = throwError $ "Read index out of bound (" ++ show idx ++ ")"
+            | fromIntegral idx >= maxInput = throwE $ "Read index out of bound (" ++ show idx ++ ")"
             | otherwise = pure $ L.index inputData (fromIntegral idx)
 
         maxOutput = M.length scanline
@@ -223,7 +245,7 @@ newStyleRLE inputData initialIdx scanline = foldM inner initialIdx [0 .. 3]
 
 
         strideSet count destIndex _ | endIndex > maxOutput + stride =
-          throwError $ "Out of bound HDR scanline " ++ show endIndex ++ " (max " ++ show maxOutput ++ ")"
+          throwE $ "Out of bound HDR scanline " ++ show endIndex ++ " (max " ++ show maxOutput ++ ")"
             where endIndex = destIndex + count * stride
         strideSet count destIndex val = aux destIndex count
             where aux i 0 =  pure i
@@ -233,7 +255,7 @@ newStyleRLE inputData initialIdx scanline = foldM inner initialIdx [0 .. 3]
 
 
         strideCopy _ count destIndex
-            | writeEndBound > maxOutput + stride = throwError "Out of bound HDR scanline"
+            | writeEndBound > maxOutput + stride = throwE "Out of bound HDR scanline"
                 where writeEndBound = destIndex + count * stride
         strideCopy sourceIndex count destIndex = aux sourceIndex destIndex count
           where aux _ j 0 = pure j
@@ -369,10 +391,44 @@ encodeScanlineColor vec outVec outIdx = do
 writeHDR :: FilePath -> Image PixelRGBF -> IO ()
 writeHDR filename img = L.writeFile filename $ encodeHDR img
 
+-- | Write a RLE encoded High dynamic range image into a radiance
+-- image file on disk.
+writeRLENewStyleHDR :: FilePath -> Image PixelRGBF -> IO ()
+writeRLENewStyleHDR filename img =
+    L.writeFile filename $ encodeRLENewStyleHDR img
+
 -- | Encode an High dynamic range image into a radiance image
 -- file format.
+-- Alias for encodeRawHDR
 encodeHDR :: Image PixelRGBF -> L.ByteString
-encodeHDR pic = encode $ runST $ do
+encodeHDR = encodeRawHDR
+
+-- | Encode an High dynamic range image into a radiance image
+-- file format. without compression
+encodeRawHDR :: Image PixelRGBF -> L.ByteString
+encodeRawHDR pic = encode descriptor
+  where
+    newImage = pixelMap rgbeInRgba pic
+    -- we are cheating to death here, the layout we want
+    -- correspond to the layout of pixelRGBA8, so we
+    -- convert
+    rgbeInRgba pixel = PixelRGBA8 r g b e
+      where RGBE r g b e = toRGBE pixel
+
+    descriptor = RadianceHeader
+        { radianceInfos = []
+        , radianceFormat = FormatRGBE
+        , radianceHeight = imageHeight pic
+        , radianceWidth  = imageWidth pic
+        , radianceData = L.fromChunks [toByteString $ imageData newImage]
+        }
+
+
+-- | Encode an High dynamic range image into a radiance image
+-- file format using a light RLE compression. Some problems
+-- seem to arise with some image viewer.
+encodeRLENewStyleHDR :: Image PixelRGBF -> L.ByteString
+encodeRLENewStyleHDR pic = encode $ runST $ do
     let w = imageWidth pic
         h = imageHeight pic
 
@@ -443,7 +499,7 @@ decodeRadiancePicture hdr = do
               inner | isNewRunLengthMarker color = do
                           let calcSize = checkLineLength color
                           when (calcSize /= width)
-                               (throwError "Invalid sanline size")
+                               (throwE "Invalid sanline size")
                           pure $ \idx -> newStyleRLE packedData (idx + 4)
                     | otherwise = pure $ oldStyleRLE packedData
           f <- inner
