@@ -3,6 +3,7 @@
 {-# LANGUAGE ScopedTypeVariables #-}
 {-# LANGUAGE FlexibleInstances #-}
 {-# LANGUAGE BangPatterns #-}
+{-# LANGUAGE TypeFamilies #-}
 {-# OPTIONS_GHC -fspec-constr-count=5 #-}
 -- | Module used for JPEG file loading and writing.
 module Codec.Picture.Jpg( decodeJpeg
@@ -476,10 +477,11 @@ decodeJpeg file = case runGetStrict get file of
             imgHeight = fromIntegral $ jpgHeight scanInfo
 
             imageSize = imgWidth * imgHeight * compCount
-            (st, wrotten) = execRWS (mapM_ jpgMachineStep (jpgFrame img)) () emptyDecoderState
-            Just fHdr = currentFrame st
 
-            decodeProgressive = runST $
+            decodeProgressive = runST $ do
+                let (st, wrotten) =
+                        execRWS (mapM_ jpgMachineStep (jpgFrame img)) () emptyDecoderState
+                    Just fHdr = currentFrame st
                 progressiveUnpack
                     (maximumHorizontalResolution st, maximumVerticalResolution st)
                     fHdr
@@ -487,6 +489,9 @@ decodeJpeg file = case runGetStrict get file of
                     wrotten >>= unsafeFreezeImage
 
             pixelData = runST $ do
+                let (st, wrotten) =
+                        execRWS (mapM_ jpgMachineStep (jpgFrame img)) () emptyDecoderState
+                    Just fHdr = currentFrame st
                 resultImage <- M.new imageSize
                 let wrapped = MutableImage imgWidth imgHeight resultImage
                 decodeImage 
